# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import TYPE_CHECKING

from faebryk.core.moduleinterface import ModuleInterface

if TYPE_CHECKING:
    from faebryk.library.Pad import Pad


class has_linked_pad(ModuleInterface.TraitT):
    @abstractmethod
    def get_pads(self) -> set["Pad"]: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path

import typer

from faebryk.libs.tools.typer import typer_callback

TEMPLATE_REPO_URL = "https://github.com/faebryk/project-template.git"


@dataclass
class CTX:
    name: str
    template_repo_url: str
    branch: str
    target_dir: Path
    force_overwrite: bool


@typer_callback(None)
def main(
    ctx: typer.Context,
    name: str,
    template_repo_url: str = TEMPLATE_REPO_URL,
    branch: str = "main",
    target_dir: Path = Path.cwd(),
    force_overwrite: bool = False,
):
    """
    Can be called like this: > faebryk project
    Or python -m faebryk project
    Or python -m faebryk.tools.project
    For help invoke command without arguments.
    """

    ctx.obj = CTX(
        name,
        template_repo_url,
        branch,
        target_dir,
        force_overwrite,
    )


# TODO add remote option
#   to do `gh repo create --template <template> <name>`


@main.command()
def local(ctx: typer.Context, cache: bool = True):
    obj = ctx.obj

    target: Path = obj.target_dir / obj.name

    if target.exists():
        if obj.force_overwrite:
            typer.echo(f"Directory {target} already exists. Overwriting.")
            subprocess.check_output(["rm", "-rf", str(target)])
        else:
            typer.echo(f"Directory {target} already exists. Aborting.")
            raise typer.Exit(code=1)

    # git clone
    subprocess.check_output(
        [
            "git",
            "clone",
            "--branch",
            obj.branch,
            "--",
            obj.template_repo_url,
            str(target),
        ]
    )

    # setup project
    p = subprocess.Popen(
        [
            sys.executable,
            target / "scripts/setup_project.py",
            "--no-cache" if not cache else "--cache",
        ]
    )
    p.wait()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from pathlib import Path

from faebryk.libs.util import (
    ConfigFlag,
    FileChangedWatcher,
    at_exit,
    find_file,
    global_lock,
    is_editable_install,
    least_recently_modified_file,
)

logger = logging.getLogger(__name__)


LEAK_WARNINGS = ConfigFlag("CPP_LEAK_WARNINGS", default=False)
DEBUG_BUILD = ConfigFlag("CPP_DEBUG_BUILD", default=False)
PRINTF_DEBUG = ConfigFlag("CPP_PRINTF_DEBUG", default=False)

EDITABLE_MODULE_NAME = "faebryk_core_cpp_editable"

_thisfile = Path(__file__)
_thisdir = _thisfile.parent
_cmake_dir = _thisdir
_build_dir = _cmake_dir / "build"
_pyi_source = _build_dir / f"{EDITABLE_MODULE_NAME}.pyi"
_src_files = [_thisdir / "src", _thisdir / "include", _thisdir / "CMakeLists.txt"]


def compile():
    import platform
    import shutil
    import sys

    from faebryk.libs.header import formatted_file_contents, get_header
    from faebryk.libs.util import run_live

    logger.warning("Recompiling C++ code")

    pyi_watcher = FileChangedWatcher(_pyi_source, FileChangedWatcher.CheckMethod.HASH)

    # check for cmake binary existing
    if not shutil.which("cmake"):
        raise RuntimeError(
            "cmake not found, needed for compiling c++ code in editable mode"
        )

    other_flags = []

    # On OSx we've had some issues with building for the right architecture
    if sys.platform == "darwin":  # macOS
        arch = platform.machine()
        if arch in ["arm64", "x86_64"]:
            other_flags += [f"-DCMAKE_OSX_ARCHITECTURES={arch}"]

    if DEBUG_BUILD:
        other_flags += ["-DCMAKE_BUILD_TYPE=Debug"]
    other_flags += [f"-DGLOBAL_PRINTF_DEBUG={int(bool(PRINTF_DEBUG))}"]

    run_parallel = ["-j"]
    if sys.platform == "win32":
        # Use /m for parallel builds with MSVC
        run_parallel = ["/m"]

    with global_lock(_build_dir / "lock", timeout_s=60):
        run_live(
            [
                "cmake",
                "-S",
                str(_cmake_dir),
                "-B",
                str(_build_dir),
                "-DEDITABLE=1",
                "-DPython_EXECUTABLE=" + sys.executable,
                *other_flags,
            ],
            stdout=logger.debug,
            stderr=logger.error,
        )
        run_live(
            [
                "cmake",
                "--build",
                str(_build_dir),
                "--",
                *run_parallel,
            ],
            stdout=logger.debug,
            stderr=logger.error,
        )

    # move autogenerated type stub file to source directory
    if pyi_watcher.has_changed():
        pyi_out = _thisfile.with_suffix(".pyi")
        pyi_out.write_text(
            formatted_file_contents(
                get_header()
                + "\n"
                + "# This file is auto-generated by nanobind.\n"
                + "# Do not edit this file directly; edit the corresponding\n"
                + "# C++ file instead.\n"
                # + "from typing import overload\n"
                + _pyi_source.read_text(encoding="utf-8"),
                is_pyi=True,
            ),
            encoding="utf-8",
        )
        run_live(
            [sys.executable, "-m", "ruff", "check", "--fix", pyi_out],
            stdout=logger.debug,
            stderr=logger.error,
        )


def load():
    import sys

    # Search the module in the build directory and its subdirectories
    if not _build_dir.exists():
        raise RuntimeError("build directory not found")

    module_dir = _build_dir
    if sys.platform == "win32":
        # Find the actual directory containing the compiled module (.pyd on Windows)
        # Check common locations: build dir itself, Debug, Release subdirs
        module_name_start = EDITABLE_MODULE_NAME
        search_paths = [_build_dir, _build_dir / "Debug", _build_dir / "Release"]
        module_path = None
        for potential_dir in search_paths:
            module_path = next(
                find_file(potential_dir, f"{module_name_start}*.pyd"), None
            )
            if module_path:
                module_dir = module_path.parent
                break
        if not module_path:
            raise RuntimeError(
                f"Could not find {module_name_start}*.pyd in "
                "build directory or subdirectories."
            )

    # Add the directory containing the .pyd file to sys.path
    sys.path.append(str(module_dir))


def compile_and_load():
    """
    Forces C++ to compile into faebryk_core_cpp_editable module which is then loaded
    into _cpp.
    """

    # Force recompile
    # subprocess.run(["rm", "-rf", str(build_dir)], check=True)

    # check if source files are newer than build files
    build_date_info = least_recently_modified_file(_build_dir)
    src_date_info = least_recently_modified_file(*_src_files)
    if (
        build_date_info is None
        or src_date_info is None
        or build_date_info[1] < src_date_info[1]
    ):
        compile()

    # load into sys.path
    load()


# Re-export c++ with type hints provided by __init__.pyi
if is_editable_install():
    compile_and_load()
    from faebryk_core_cpp_editable import *  # type: ignore # noqa: E402, F403
else:
    from faebryk_core_cpp import *  # type: ignore # noqa: E402, F403


def cleanup():
    if LEAK_WARNINGS:
        print("\n--- Nanobind leakage analysis ".ljust(80, "-"))
        # nanobind automatically prints leaked objects at exit
    from faebryk.core.cpp import set_leak_warnings

    set_leak_warnings(bool(LEAK_WARNINGS))


at_exit(cleanup, on_exception=False)



import pytest

from faebryk.core.node import FieldError, Node
from faebryk.core.reference import Reference


def test_points_to_correct_node():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

    a = A()
    b = B()
    b.x = a
    assert b.x is a


def test_immutable():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

    b = B()
    a = A()
    b.x = a

    with pytest.raises(TypeError):
        b.x = A()


def test_unset():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

    b = B()
    with pytest.raises(Reference.UnboundError):
        b.x


def test_wrong_type():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

    b = B()
    with pytest.raises(TypeError):
        b.x = 1


def test_set_value_before_constuction():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

        def __init__(self, x):
            super().__init__()
            self.x = x

    a = A()
    b = B(a)
    assert b.x is a


def test_get_value_before_constuction():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)
        y = Reference(A)

        def __init__(self, x):
            super().__init__()
            self.x = x
            self.y = self.x

    a = A()
    b = B(a)
    assert b.y is a


def test_typed_construction_doesnt_work():
    class B(Node):
        x: Reference

    # check using the property directly that everything is working
    with pytest.raises(AttributeError):
        B.x


def test_typed_construction_protection():
    """
    Ensure references aren't constructed as a field

    If properties are constructed as instance fields, their
    getters and setters aren't called when assigning to them.

    This means we won't actually construct the underlying graph properly.
    It's pretty insidious because it's completely non-obvious that we're
    missing these graph connections.
    """

    class A(Node):
        pass

    class B(Node):
        x: Reference[A]

    with pytest.raises(FieldError):
        B()


def test_underlying_property_explicitly():
    class A(Node):
        pass

    class B(Node):
        x = Reference(A)

    a = A()
    b = B()
    b.x = a

    # check using the property directly that everything is working
    assert B.x.gifs[b].get_reference() is a

    # check that the property is set
    assert b.x is a



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F


class has_kicad_manual_footprint(F.has_kicad_footprint.impl()):
    def __init__(self, str, pinmap: dict[F.Pad, str]) -> None:
        super().__init__()
        self.str = str
        self.pinmap = pinmap

    def get_kicad_footprint(self):
        return self.str

    def get_pin_names(self):
        return self.pinmap



"""
Utils related to handling the parse tree
"""

from io import StringIO
from pathlib import Path
from typing import Generator

import pygments.lexer
from antlr4 import CommonTokenStream, InputStream, ParserRuleContext, Token
from antlr4.TokenStreamRewriter import TokenStreamRewriter
from pygments import token as pygments_token

from atopile.parser.AtoLexer import AtoLexer


def get_src_info_from_token(token: Token) -> tuple[str, int, int]:
    """Get the source path, line, and column from a context"""
    input_stream: InputStream = token.getInputStream()
    return input_stream.name, token.line, token.column


def get_src_info_from_ctx(
    ctx: ParserRuleContext,
) -> tuple[str | Path, int, int, int, int]:
    """Get the source path, line, and column from a context"""
    token: Token = ctx.start
    return (
        *get_src_info_from_token(token),
        ctx.stop.line,
        ctx.stop.column + len(ctx.stop.text),
    )


def format_src_info(ctx: ParserRuleContext) -> str:
    """Format the source path, line, and column"""
    src, start_line, start_col, _, _ = get_src_info_from_ctx(ctx)
    return f"{src}:{start_line}:{start_col}"


def expand(
    stream: list[Token], beginning: int, step: int, count: int | None = None
) -> int:
    """
    Expand the stream by the given amount in the given direction
    """
    if count is None:
        return beginning

    i = beginning
    newlines_to_find = count
    while 0 <= i < len(stream):
        if stream[i].type == AtoLexer.NEWLINE:
            if newlines_to_find == 0:
                return i - step
            newlines_to_find -= 1
        i += step

    return min(max(i, 0), len(stream) - 1)


class IterableRewriter(TokenStreamRewriter):
    def iter_tokens_and_text(
        self, program_name, start: int | None = None, stop: int | None = None
    ) -> Generator[tuple[Token | None, str], None, None]:
        """
        :return: the text in tokens[start, stop](closed interval)
        """
        rewrites = self.programs.get(program_name)

        # ensure start/end are in range
        if stop is None or stop > len(self.tokens.tokens) - 1:
            stop = len(self.tokens.tokens) - 1
        if start is None or start < 0:
            start = 0

        indexToOp = self._reduceToSingleOperationPerIndex(rewrites)
        i = start
        while all((i <= stop, i < len(self.tokens.tokens))):
            op = indexToOp.pop(i, None)
            token = self.tokens.get(i)
            if op is None:
                if token.type != Token.EOF:
                    yield token, token.text
                i += 1
            else:
                buf = StringIO()
                i = op.execute(buf)
                yield token, buf.getvalue()

        if stop == len(self.tokens.tokens) - 1:
            for op in indexToOp.values():
                if op.index >= len(self.tokens.tokens) - 1:
                    yield None, op.text


class AtoRewriter(IterableRewriter):
    DROP_TOKENS = {AtoLexer.INDENT, AtoLexer.DEDENT}

    def iter_tokens_and_text(
        self, program_name, start: int | None = None, stop: int | None = None
    ) -> Generator[tuple[Token | None, str], None, None]:
        for token, text in super().iter_tokens_and_text(program_name, start, stop):
            if token and token.type in self.DROP_TOKENS:
                continue
            elif token and token.type == AtoLexer.NEWLINE:
                # FIXME: feels like we shouldn't have this
                yield token, "\n"
            else:
                yield token, text

    def getText(self, program_name, start: int, stop: int):
        return self.get_text(program_name, start, stop)

    def get_text(self, program_name, start: int, stop: int):
        return "".join(
            text for _, text in self.iter_tokens_and_text(program_name, start, stop)
        )


class PygmentsLexer(pygments.lexer.Lexer):
    TOKEN_TYPE_MAP = {
        AtoLexer.INDENT: pygments_token.Token,
        AtoLexer.DEDENT: pygments_token.Token,
        AtoLexer.STRING: pygments_token.String,
        AtoLexer.NUMBER: pygments_token.Number,
        AtoLexer.INTEGER: pygments_token.Number,
        AtoLexer.COMPONENT: pygments_token.Keyword,
        AtoLexer.MODULE: pygments_token.Keyword,
        AtoLexer.INTERFACE: pygments_token.Keyword,
        AtoLexer.PIN: pygments_token.Keyword,
        AtoLexer.SIGNAL: pygments_token.Keyword,
        AtoLexer.NEW: pygments_token.Keyword,
        AtoLexer.FROM: pygments_token.Keyword,
        AtoLexer.IMPORT: pygments_token.Keyword,
        AtoLexer.ASSERT: pygments_token.Keyword,
        AtoLexer.TO: pygments_token.Keyword,
        AtoLexer.TRUE: pygments_token.Name,
        AtoLexer.FALSE: pygments_token.Name,
        AtoLexer.WITHIN: pygments_token.Name,
        AtoLexer.PASS: pygments_token.Keyword,
        AtoLexer.NAME: pygments_token.Name,
        AtoLexer.STRING_LITERAL: pygments_token.String,
        AtoLexer.BYTES_LITERAL: pygments_token.String,
        AtoLexer.DECIMAL_INTEGER: pygments_token.Number,
        AtoLexer.OCT_INTEGER: pygments_token.Number,
        AtoLexer.HEX_INTEGER: pygments_token.Number,
        AtoLexer.BIN_INTEGER: pygments_token.Number,
        AtoLexer.FLOAT_NUMBER: pygments_token.Number,
        AtoLexer.IMAG_NUMBER: pygments_token.Number,
        AtoLexer.PLUS_OR_MINUS: pygments_token.Operator,
        AtoLexer.PLUS_SLASH_MINUS: pygments_token.Operator,
        AtoLexer.PLUS_MINUS_SIGN: pygments_token.Operator,
        AtoLexer.PERCENT: pygments_token.Operator,
        AtoLexer.DOT: pygments_token.Token,
        AtoLexer.ELLIPSIS: pygments_token.Token,
        AtoLexer.STAR: pygments_token.Token,
        AtoLexer.OPEN_PAREN: pygments_token.Token,
        AtoLexer.CLOSE_PAREN: pygments_token.Token,
        AtoLexer.COMMA: pygments_token.Token,
        AtoLexer.COLON: pygments_token.Token,
        AtoLexer.SEMI_COLON: pygments_token.Token,
        AtoLexer.POWER: pygments_token.Token,
        AtoLexer.ASSIGN: pygments_token.Operator,
        AtoLexer.OPEN_BRACK: pygments_token.Token,
        AtoLexer.CLOSE_BRACK: pygments_token.Token,
        AtoLexer.OR_OP: pygments_token.Token,
        AtoLexer.XOR: pygments_token.Token,
        AtoLexer.AND_OP: pygments_token.Token,
        AtoLexer.LEFT_SHIFT: pygments_token.Token,
        AtoLexer.RIGHT_SHIFT: pygments_token.Token,
        AtoLexer.PLUS: pygments_token.Token,
        AtoLexer.MINUS: pygments_token.Token,
        AtoLexer.DIV: pygments_token.Token,
        AtoLexer.IDIV: pygments_token.Token,
        AtoLexer.SPERM: pygments_token.Token,
        AtoLexer.WIRE: pygments_token.Token,
        AtoLexer.OPEN_BRACE: pygments_token.Token,
        AtoLexer.CLOSE_BRACE: pygments_token.Token,
        AtoLexer.LESS_THAN: pygments_token.Operator,
        AtoLexer.GREATER_THAN: pygments_token.Operator,
        AtoLexer.EQUALS: pygments_token.Operator,
        AtoLexer.GT_EQ: pygments_token.Operator,
        AtoLexer.LT_EQ: pygments_token.Operator,
        AtoLexer.NOT_EQ_1: pygments_token.Operator,
        AtoLexer.NOT_EQ_2: pygments_token.Operator,
        AtoLexer.AT: pygments_token.Token,
        AtoLexer.ARROW: pygments_token.Token,
        AtoLexer.ADD_ASSIGN: pygments_token.Token,
        AtoLexer.SUB_ASSIGN: pygments_token.Token,
        AtoLexer.MULT_ASSIGN: pygments_token.Token,
        AtoLexer.AT_ASSIGN: pygments_token.Token,
        AtoLexer.DIV_ASSIGN: pygments_token.Token,
        AtoLexer.AND_ASSIGN: pygments_token.Token,
        AtoLexer.OR_ASSIGN: pygments_token.Token,
        AtoLexer.XOR_ASSIGN: pygments_token.Token,
        AtoLexer.LEFT_SHIFT_ASSIGN: pygments_token.Token,
        AtoLexer.RIGHT_SHIFT_ASSIGN: pygments_token.Token,
        AtoLexer.POWER_ASSIGN: pygments_token.Token,
        AtoLexer.IDIV_ASSIGN: pygments_token.Token,
        AtoLexer.NEWLINE: pygments_token.Whitespace,
        AtoLexer.COMMENT: pygments_token.Comment,
        AtoLexer.WS: pygments_token.Whitespace,
        AtoLexer.EXPLICIT_LINE_JOINING: pygments_token.Token,
        AtoLexer.ERRORTOKEN: pygments_token.Error,
        AtoLexer.FOR: pygments_token.Keyword,
        AtoLexer.IN: pygments_token.Keyword,
    }

    name = "atopile"

    aliases = ["ato"]

    filenames = ["*.ato"]

    def get_tokens_unprocessed(self, text: str):
        text_stream = InputStream(text)
        lexer = AtoLexer(text_stream)
        token_stream = CommonTokenStream(lexer)
        token_stream.fill()
        rewriter = AtoRewriter(token_stream)

        for token, text in rewriter.iter_tokens_and_text(
            TokenStreamRewriter.DEFAULT_PROGRAM_NAME
        ):
            if token is not None:
                yield (
                    token.start,
                    PygmentsLexer.TOKEN_TYPE_MAP.get(token.type, pygments_token.Token),  # type: ignore  # obviously random other shit is why I'm using .get
                    text,
                )


class PygmentsLexerReconstructor(pygments.lexer.Lexer):
    """Reconstruct source-code from a ctx, and antlr4 rewriter"""

    def __init__(
        self,
        rewriter: AtoRewriter,
        start: int,
        stop: int,
    ):
        self.rewriter = rewriter
        self.start = start
        self.stop = stop

    def get_code(self) -> str:
        return self.rewriter.getText(
            TokenStreamRewriter.DEFAULT_PROGRAM_NAME, self.start, self.stop
        )

    def get_tokens(
        self, _: str
    ) -> Generator[tuple["pygments_token._TokenType", str], None, None]:
        for token, text in self.rewriter.iter_tokens_and_text(
            TokenStreamRewriter.DEFAULT_PROGRAM_NAME, self.start, self.stop
        ):
            yield (
                PygmentsLexer.TOKEN_TYPE_MAP.get(token.type, pygments_token.Token),  # type: ignore  # obviously random other shit is why I'm using .get
                text,
            )

    @classmethod
    def from_tokens(
        cls,
        token_stream: CommonTokenStream,
        start_token: Token,
        stop_token: Token | None = None,
        expand_before: int | None = None,
        expand_after: int | None = None,
    ) -> "PygmentsLexerReconstructor":
        rewriter = AtoRewriter(token_stream)

        start_index = expand(
            token_stream.tokens, start_token.tokenIndex, -1, expand_before
        )

        if stop_token is None:
            stop_token = start_token

        stop_index = expand(token_stream.tokens, stop_token.tokenIndex, 1, expand_after)

        return PygmentsLexerReconstructor(rewriter, start_index, stop_index)

    @classmethod
    def from_ctx(
        cls,
        ctx: ParserRuleContext,
        expand_before: int | None = None,
        expand_after: int | None = None,
    ) -> "PygmentsLexerReconstructor":
        return cls.from_tokens(
            ctx.parser.getInputStream(),  # type: ignore
            ctx.start,
            ctx.stop,
            expand_before,
            expand_after,
        )

    @property
    def start_line(self) -> int:
        return self.rewriter.tokens.tokens[self.start].line


def reconstruct(
    ctx: ParserRuleContext,
    expand_before: int | None = None,
    expand_after: int | None = None,
) -> str:
    """
    Reconstruct the source code from a parse tree

    Args:
        ctx: The parse tree to reconstruct
        expand_before: The number of lines to expand before the mark.
            0 -> To the start of the current line
        expand_after: The number of lines to expand after the mark.
            0 -> To the end of the current line
    """
    return PygmentsLexerReconstructor.from_ctx(
        ctx, expand_before, expand_after
    ).get_code()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P, Quantity
from faebryk.libs.util import times


class QFN(F.Footprint):
    def __init__(
        self,
        pin_cnt: int,
        exposed_thermal_pad_cnt: int,
        size_xy: tuple[Quantity, Quantity],
        pitch: Quantity,
        exposed_thermal_pad_dimensions: tuple[Quantity, Quantity],
        has_thermal_vias: bool,
    ) -> None:
        super().__init__()

        self._pin_cnt = pin_cnt
        self._exposed_thermal_pad_cnt = exposed_thermal_pad_cnt
        self._size_xy = size_xy
        self._pitch = pitch
        self._exposed_thermal_pad_dimensions = exposed_thermal_pad_dimensions
        self._has_thermal_vias = has_thermal_vias

        assert exposed_thermal_pad_cnt > 0 or not has_thermal_vias
        assert (
            exposed_thermal_pad_dimensions[0] < size_xy[0]
            and exposed_thermal_pad_dimensions[1] < size_xy[1]
        )

    @L.rt_field
    def pins(self):
        return times(self._pin_cnt, F.Pad)

    equal_pins: F.has_equal_pins_in_ifs
    attach_via_pinmap: F.can_attach_via_pinmap_equal

    @L.rt_field
    def kicad_footprint(self):
        class _has_kicad_footprint(F.has_kicad_footprint_equal_ifs):
            @staticmethod
            def get_kicad_footprint() -> str:
                return "Package_DFN_QFN:QFN-{leads}-{ep}EP_{size_x}x{size_y}mm_P{pitch}mm_EP{ep_x}x{ep_y}mm{vias}".format(  # noqa: E501
                    leads=self._pin_cnt,
                    ep=self._exposed_thermal_pad_cnt,
                    size_x=self._size_xy[0].to(P.mm).m,
                    size_y=self._size_xy[1].to(P.mm).m,
                    pitch=self._pitch.to(P.mm).m,
                    ep_x=self._exposed_thermal_pad_dimensions[0].to(P.mm).m,
                    ep_y=self._exposed_thermal_pad_dimensions[1].to(P.mm).m,
                    vias="_ThermalVias" if self._has_thermal_vias else "",
                )

        return _has_kicad_footprint()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class ElecSuper_PSM712_ES(Module):
    """
    RS485 bus ESD and surge protection

    17A 350W Bidirectional SOT-23
    ESD and Surge Protection (TVS/ESD) ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    rs485: F.RS485HalfDuplex

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    explicit_part = L.f_field(F.has_explicit_part.by_mfr)("ElecSuper", "PSM712-ES")

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.rs485.diff_pair.n.line: ["1"],
                self.rs485.diff_pair.p.line: ["2"],
                self.rs485.diff_pair.n.reference.lv: ["3"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



# Generated from AtoParser.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .AtoParser import AtoParser
else:
    from AtoParser import AtoParser

# This class defines a complete generic visitor for a parse tree produced by AtoParser.

class AtoParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by AtoParser#file_input.
    def visitFile_input(self, ctx:AtoParser.File_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pragma_stmt.
    def visitPragma_stmt(self, ctx:AtoParser.Pragma_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#stmt.
    def visitStmt(self, ctx:AtoParser.StmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#simple_stmts.
    def visitSimple_stmts(self, ctx:AtoParser.Simple_stmtsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#simple_stmt.
    def visitSimple_stmt(self, ctx:AtoParser.Simple_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#compound_stmt.
    def visitCompound_stmt(self, ctx:AtoParser.Compound_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blockdef.
    def visitBlockdef(self, ctx:AtoParser.BlockdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blockdef_super.
    def visitBlockdef_super(self, ctx:AtoParser.Blockdef_superContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blocktype.
    def visitBlocktype(self, ctx:AtoParser.BlocktypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#block.
    def visitBlock(self, ctx:AtoParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#dep_import_stmt.
    def visitDep_import_stmt(self, ctx:AtoParser.Dep_import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#import_stmt.
    def visitImport_stmt(self, ctx:AtoParser.Import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#declaration_stmt.
    def visitDeclaration_stmt(self, ctx:AtoParser.Declaration_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#field_reference_or_declaration.
    def visitField_reference_or_declaration(self, ctx:AtoParser.Field_reference_or_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assign_stmt.
    def visitAssign_stmt(self, ctx:AtoParser.Assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_assign_stmt.
    def visitCum_assign_stmt(self, ctx:AtoParser.Cum_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#set_assign_stmt.
    def visitSet_assign_stmt(self, ctx:AtoParser.Set_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_operator.
    def visitCum_operator(self, ctx:AtoParser.Cum_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_assignable.
    def visitCum_assignable(self, ctx:AtoParser.Cum_assignableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assignable.
    def visitAssignable(self, ctx:AtoParser.AssignableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#retype_stmt.
    def visitRetype_stmt(self, ctx:AtoParser.Retype_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#directed_connect_stmt.
    def visitDirected_connect_stmt(self, ctx:AtoParser.Directed_connect_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#connect_stmt.
    def visitConnect_stmt(self, ctx:AtoParser.Connect_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bridgeable.
    def visitBridgeable(self, ctx:AtoParser.BridgeableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#mif.
    def visitMif(self, ctx:AtoParser.MifContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#connectable.
    def visitConnectable(self, ctx:AtoParser.ConnectableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#signaldef_stmt.
    def visitSignaldef_stmt(self, ctx:AtoParser.Signaldef_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pindef_stmt.
    def visitPindef_stmt(self, ctx:AtoParser.Pindef_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pin_declaration.
    def visitPin_declaration(self, ctx:AtoParser.Pin_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pin_stmt.
    def visitPin_stmt(self, ctx:AtoParser.Pin_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#new_stmt.
    def visitNew_stmt(self, ctx:AtoParser.New_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#new_count.
    def visitNew_count(self, ctx:AtoParser.New_countContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#string_stmt.
    def visitString_stmt(self, ctx:AtoParser.String_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pass_stmt.
    def visitPass_stmt(self, ctx:AtoParser.Pass_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#list_literal_of_field_references.
    def visitList_literal_of_field_references(self, ctx:AtoParser.List_literal_of_field_referencesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#iterable_references.
    def visitIterable_references(self, ctx:AtoParser.Iterable_referencesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#for_stmt.
    def visitFor_stmt(self, ctx:AtoParser.For_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assert_stmt.
    def visitAssert_stmt(self, ctx:AtoParser.Assert_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#trait_stmt.
    def visitTrait_stmt(self, ctx:AtoParser.Trait_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#constructor.
    def visitConstructor(self, ctx:AtoParser.ConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#template.
    def visitTemplate(self, ctx:AtoParser.TemplateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#template_arg.
    def visitTemplate_arg(self, ctx:AtoParser.Template_argContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#comparison.
    def visitComparison(self, ctx:AtoParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#compare_op_pair.
    def visitCompare_op_pair(self, ctx:AtoParser.Compare_op_pairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#lt_arithmetic_or.
    def visitLt_arithmetic_or(self, ctx:AtoParser.Lt_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#gt_arithmetic_or.
    def visitGt_arithmetic_or(self, ctx:AtoParser.Gt_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#lt_eq_arithmetic_or.
    def visitLt_eq_arithmetic_or(self, ctx:AtoParser.Lt_eq_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#gt_eq_arithmetic_or.
    def visitGt_eq_arithmetic_or(self, ctx:AtoParser.Gt_eq_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#in_arithmetic_or.
    def visitIn_arithmetic_or(self, ctx:AtoParser.In_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#is_arithmetic_or.
    def visitIs_arithmetic_or(self, ctx:AtoParser.Is_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#arithmetic_expression.
    def visitArithmetic_expression(self, ctx:AtoParser.Arithmetic_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#sum.
    def visitSum(self, ctx:AtoParser.SumContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#term.
    def visitTerm(self, ctx:AtoParser.TermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#power.
    def visitPower(self, ctx:AtoParser.PowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#functional.
    def visitFunctional(self, ctx:AtoParser.FunctionalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bound.
    def visitBound(self, ctx:AtoParser.BoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#slice.
    def visitSlice(self, ctx:AtoParser.SliceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#slice_start.
    def visitSlice_start(self, ctx:AtoParser.Slice_startContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#slice_stop.
    def visitSlice_stop(self, ctx:AtoParser.Slice_stopContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#slice_step.
    def visitSlice_step(self, ctx:AtoParser.Slice_stepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#atom.
    def visitAtom(self, ctx:AtoParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#arithmetic_group.
    def visitArithmetic_group(self, ctx:AtoParser.Arithmetic_groupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#literal_physical.
    def visitLiteral_physical(self, ctx:AtoParser.Literal_physicalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bound_quantity.
    def visitBound_quantity(self, ctx:AtoParser.Bound_quantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bilateral_quantity.
    def visitBilateral_quantity(self, ctx:AtoParser.Bilateral_quantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#quantity.
    def visitQuantity(self, ctx:AtoParser.QuantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bilateral_tolerance.
    def visitBilateral_tolerance(self, ctx:AtoParser.Bilateral_toleranceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#key.
    def visitKey(self, ctx:AtoParser.KeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#array_index.
    def visitArray_index(self, ctx:AtoParser.Array_indexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pin_reference_end.
    def visitPin_reference_end(self, ctx:AtoParser.Pin_reference_endContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#field_reference_part.
    def visitField_reference_part(self, ctx:AtoParser.Field_reference_partContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#field_reference.
    def visitField_reference(self, ctx:AtoParser.Field_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#type_reference.
    def visitType_reference(self, ctx:AtoParser.Type_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#unit.
    def visitUnit(self, ctx:AtoParser.UnitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#type_info.
    def visitType_info(self, ctx:AtoParser.Type_infoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#name.
    def visitName(self, ctx:AtoParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#literal.
    def visitLiteral(self, ctx:AtoParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#string.
    def visitString(self, ctx:AtoParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#boolean_.
    def visitBoolean_(self, ctx:AtoParser.Boolean_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#number_hint_natural.
    def visitNumber_hint_natural(self, ctx:AtoParser.Number_hint_naturalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#number_hint_integer.
    def visitNumber_hint_integer(self, ctx:AtoParser.Number_hint_integerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#number.
    def visitNumber(self, ctx:AtoParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#number_signless.
    def visitNumber_signless(self, ctx:AtoParser.Number_signlessContext):
        return self.visitChildren(ctx)



del AtoParser


# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class pf_533984002(Module):
    # interfaces
    pin = L.list_field(2, F.Electrical)
    mount = L.list_field(2, F.Electrical)

    @L.rt_field
    def attach_to_footprint(self):
        x = self
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": x.pin[0],
                "2": x.pin[1],
                "3": x.mount[0],
                "4": x.mount[1],
            }
        )

    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://wmsc.lcsc.com/wmsc/upload/file/pdf/v2/lcsc/1912111437_SHOU-HAN-1-25-2P_C393945.pdf"
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import itertools
import logging

import faebryk.library._F as F
from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.module import Module
from faebryk.exporters.netlist.graph import can_represent_kicad_footprint
from faebryk.exporters.netlist.netlist import FBRKNetlist
from faebryk.libs.kicad.fileformats_latest import C_fields, C_kicad_netlist_file
from faebryk.libs.util import duplicates

logger = logging.getLogger(__name__)


def faebryk_netlist_to_kicad(fbrk_netlist: FBRKNetlist):
    tstamp = itertools.count(1)
    net_code = itertools.count(1)

    # KiCAD Constraints:
    #   - name has to be unique
    #   - vertex properties has to contain footprint
    #   - tstamps can be generated (unique)
    #   - net_code can be generated (ascending, continuous)
    #   - components unique

    dupes = duplicates(fbrk_netlist.comps, lambda comp: comp.name)
    assert not dupes, f"Duplicate comps {dupes}"

    NetlistFile = C_kicad_netlist_file
    Component = NetlistFile.C_netlist.C_components.C_component
    comps = [
        Component(
            ref=comp.name,
            value=comp.value,
            footprint=comp.properties["footprint"],
            propertys={
                k: Component.C_property(k, v)
                for k, v in comp.properties.items()
                if k != "footprint"
            },
            tstamps=str(next(tstamp)),
            fields=C_fields(
                {
                    k: C_fields.C_field(k, v)
                    for k, v in comp.properties.get("fields", [])
                }
            ),
        )
        # sort because tstamp determined by pos
        for comp in sorted(fbrk_netlist.comps, key=lambda comp: comp.name)
    ]

    # check if all vertices have a component in pre_comps
    # not sure if this is necessary
    pre_comp_names = {comp.name for comp in fbrk_netlist.comps}
    for net in fbrk_netlist.nets:
        for vertex in net.vertices:
            assert (
                vertex.component.name in pre_comp_names
            ), f"Missing {vertex.component}"

    Net = NetlistFile.C_netlist.C_nets.C_net
    nets = [
        Net(
            code=next(net_code),
            name=net.properties["name"],
            nodes=[
                Net.C_node(
                    ref=vertex.component.name,
                    pin=vertex.pin,
                )
                for vertex in sorted(net.vertices, key=lambda vert: vert.component.name)
            ],
        )
        # sort because code determined by pos
        for net in sorted(fbrk_netlist.nets, key=lambda net: net.properties["name"])
    ]

    return NetlistFile(
        export=NetlistFile.C_netlist(
            version="E",
            components=NetlistFile.C_netlist.C_components(comps=comps),
            nets=NetlistFile.C_netlist.C_nets(nets=nets),
        )
    )


def attach_kicad_info(G: Graph) -> None:
    """Attach kicad info to the footprints in the graph."""
    # group comps & fps
    node_fps = {
        n: t.get_footprint()
        # TODO maybe nicer to just look for footprints
        # and get their respective components instead
        for n, t in GraphFunctions(G).nodes_with_trait(F.has_footprint)
        if isinstance(n, Module)
    }

    logger.info(f"Found {len(node_fps)} components with footprints")
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug(f"node_fps: {node_fps}")

    # add trait/info to footprints
    for n, fp in node_fps.items():
        if fp.has_trait(can_represent_kicad_footprint):
            continue
        fp.add(can_represent_kicad_footprint(n, G))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import sys
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from enum import Enum, auto
from itertools import chain
from types import UnionType
from typing import Any, Callable, Iterable, Sequence, cast

from more_itertools import first
from rich.table import Table
from rich.tree import Tree

from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.parameter import (
    ConstrainableExpression,
    Domain,
    Expression,
    Is,
    IsSubset,
    Parameter,
    ParameterOperatable,
)
from faebryk.core.solver.algorithm import SolverAlgorithm
from faebryk.core.solver.utils import (
    S_LOG,
    SHOW_SS_IS,
    VERBOSE_TABLE,
    CanonicalExpression,
    ContradictionByLiteral,
    MutatorUtils,
    SolverAll,
    SolverAllExtended,
    SolverLiteral,
    get_graphs,
)
from faebryk.libs.exceptions import downgrade
from faebryk.libs.logging import rich_to_string
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval,
    Quantity_Interval_Disjoint,
    Quantity_Set,
)
from faebryk.libs.sets.sets import P_Set, as_lit
from faebryk.libs.units import HasUnit, Quantity, Unit, quantity
from faebryk.libs.util import (
    KeyErrorNotFound,
    cast_assert,
    duplicates,
    groupby,
    indented_container,
    invert_dict,
    once,
    unique_ref,
)

logger = logging.getLogger(__name__)
if S_LOG:
    logger.setLevel(logging.DEBUG)


@dataclass
class Transformations:
    input_print_context: ParameterOperatable.ReprContext

    mutated: dict[ParameterOperatable, ParameterOperatable] = field(
        default_factory=dict
    )
    removed: set[ParameterOperatable] = field(default_factory=set)
    copied: set[ParameterOperatable] = field(default_factory=set)
    created: dict[ParameterOperatable, list[ParameterOperatable]] = field(
        default_factory=lambda: defaultdict(list)
    )
    # TODO make api for contraining
    terminated: set[ConstrainableExpression] = field(default_factory=set)
    soft_replaced: dict[ParameterOperatable, ParameterOperatable] = field(
        default_factory=dict
    )

    @property
    def dirty(self) -> bool:
        non_no_op_mutations = any(
            k is not v for k, v in self.mutated.items() if k not in self.copied
        )

        return bool(
            self.removed or non_no_op_mutations or self.created or self.terminated
        )

    @property
    def is_identity(self) -> bool:
        return (
            not self.removed
            and all(k is v for k, v in self.mutated.items())
            and not self.created
            and not self.terminated
        )

    @property
    def touched_graphs(self) -> set[Graph]:
        """
        Return graphs that require a copy in some form
        - if a mutation happened we need to copy the whole graph to replace
         the old node with the new one
        - if a node was removed, we need to copy the graph to remove it
        """
        return set(get_graphs(self.removed | self.mutated.keys()))

    @staticmethod
    def identity(
        *gs: Graph, input_print_context: ParameterOperatable.ReprContext
    ) -> "Transformations":
        return Transformations(
            mutated={
                po: po for po in GraphFunctions(*gs).nodes_of_type(ParameterOperatable)
            },
            input_print_context=input_print_context,
        )

    # TODO careful with once, need to check if illegal call when not done
    @property
    @once
    def output_print_context(self) -> ParameterOperatable.ReprContext:
        context_old = self.input_print_context
        if self.is_identity:
            return context_old

        context_new = ParameterOperatable.ReprContext()
        context_new.variable_mapping.next_id = context_old.variable_mapping.next_id

        for s, d in self.mutated.items():
            if isinstance(s, Parameter) and isinstance(d, Parameter):
                s.compact_repr(context_old)
                s_mapping = context_old.variable_mapping.mapping[s]
                d_mapping = context_new.variable_mapping.mapping.get(d, None)
                if d_mapping is None or d_mapping > s_mapping:
                    context_new.variable_mapping.mapping[d] = s_mapping

        return context_new

    def __str__(self) -> str:
        if not self.dirty:
            return "Transformations()"
        assert self.input_print_context

        old_context = self.input_print_context
        new_context = self.output_print_context

        mutated_transformations = [
            (k.compact_repr(old_context), v.compact_repr(new_context))
            for k, v in self.mutated.items()
            if k not in self.copied
        ]
        mutated = indented_container(
            [f"{k} -> {v}" for k, v in mutated_transformations if k != v]
            + [f"copy {k}" for k, v in mutated_transformations if k == v]
        )
        created = indented_container(
            [k.compact_repr(new_context) for k in self.created]
        )
        removed = indented_container(
            [k.compact_repr(old_context) for k in self.removed]
        )
        # copied = indented_container(
        #    [k.compact_repr(old_context) for k in self.transformations.copied]
        # )
        copied = len(self.copied)
        terminated = len(self.terminated)
        return (
            f"mutated={mutated}"
            f", created={created}"
            f", removed={removed}"
            f", copied={copied}"
            f", terminated={terminated}"
        )

    def get_new_constraints(
        self, op: ParameterOperatable
    ) -> list[ConstrainableExpression]:
        # TODO could still happen, but then we have clash
        # keep this in mind for future
        if self.is_identity:
            return []
        if op not in self.copied:
            return []
        target = self.mutated[op]
        out = []
        for e in self.created:
            if not isinstance(e, ConstrainableExpression) or not e.constrained:
                continue
            if target in e.get_operand_operatables():
                out.append(e)
        return out


@dataclass
class Traceback:
    class Type(Enum):
        NOOP = auto()
        PASSTHROUGH = auto()
        COPIED = auto()
        CREATED = auto()
        SOFT_REPLACED = auto()
        MERGED = auto()
        MUTATED = auto()
        CONSTRAINED = auto()

    @dataclass(repr=False)
    class Stage:
        srcs: Sequence[ParameterOperatable]
        dst: ParameterOperatable
        algo: str
        reason: "Traceback.Type"
        src_context: ParameterOperatable.ReprContext
        dst_context: ParameterOperatable.ReprContext
        related: list["Traceback.Stage"]

        def __repr__(self) -> str:
            return f"{self.reason.name} {self.algo}"

    stage: Stage
    back: "list[Traceback]" = field(default_factory=list)

    def visit(self, visitor: Callable[["Traceback", int], bool]) -> None:
        """
        Visit all nodes in the traceback tree in a depth-first manner without recursion.

        Args:
            visitor: A function that takes a Traceback node and depth as arguments.
                    Returns True to continue traversal into children, False to skip.
        """
        # Stack contains tuples of (node, depth)
        stack: list[tuple[Traceback, int]] = [(self, 0)]

        while stack:
            current, depth = stack.pop()

            # Visit the current node
            continue_traversal = visitor(current, depth)

            # If visitor returns True and there are children, add them to the stack
            if continue_traversal and current.back:
                # Add children in reverse order to maintain DFS left-to-right traversal
                for child in reversed(current.back):
                    stack.append((child, depth + 1))

    def filtered(self) -> "Traceback":
        """
        NOOP & PASSTHROUGH stages always have exactly one source
            (which is the destination)
        This function returns a new traceback with all NOOP & PASSTHROUGH stages removed
        Root is always kept.
        ```
        CREATED
         NOOP
          COPIED
        ```
        becomes
        ```
        CREATED
         COPIED
        ```
        """

        # Create a mapping of original nodes to their filtered counterparts
        node_map: dict[int, Traceback] = {}

        # Create a new root traceback with the same stage as the original
        result = Traceback(stage=self.stage)
        node_map[id(self)] = result

        # Stack for DFS traversal: (original_node, filtered_parent)
        stack: list[tuple[Traceback, Traceback]] = []

        # Initialize stack with children of root
        for child in self.back:
            stack.append((child, result))

        while stack:
            original, filtered_parent = stack.pop()

            if original.stage.reason in {
                Traceback.Type.NOOP,
                Traceback.Type.PASSTHROUGH,
                Traceback.Type.COPIED,
            }:
                # For NOOP stages, skip this node but process its children
                for grandchild in original.back:
                    stack.append((grandchild, filtered_parent))
            else:
                # For non-NOOP stages, create a filtered node
                filtered_node = Traceback(stage=original.stage)
                filtered_parent.back.append(filtered_node)
                node_map[id(original)] = filtered_node

                # Process children of this node
                for child in original.back:
                    stack.append((child, filtered_node))

        return result

    def get_leaves(self) -> list[ParameterOperatable]:
        leaves = []

        def _collect_leaves(node, depth):
            if not node.back:
                leaves.extend(node.stage.srcs)
            return True

        self.visit(_collect_leaves)
        return leaves

    def __repr__(self) -> str:
        # TODO
        return f"Traceback({id(self):04x}) {self.stage}"

    def as_rich_tree(self, visited: set[ParameterOperatable] | None = None) -> Tree:
        from rich.text import Text

        if visited is None:
            visited = set()

        dst_text = self.stage.dst.compact_repr(self.stage.dst_context)
        tree = Tree(Text(dst_text, style="bold blue"))

        if self.stage.dst in visited:
            tree.add(Text("...duplicate...", style="bold red"))
            return tree

        if self.stage.reason not in {
            Traceback.Type.NOOP,
            Traceback.Type.PASSTHROUGH,
        }:
            visited.add(self.stage.dst)

        reason = self.stage.reason.name
        algo = " ".join(self.stage.algo.split(" ")[:3])

        # Create a node for the reason and algorithm
        if self.stage.reason in {
            Traceback.Type.NOOP,
            Traceback.Type.PASSTHROUGH,
            Traceback.Type.COPIED,
        }:
            reason_branch = tree
        else:
            node_text = Text(f"{reason}", style="bold cyan")
            node_text.append(f"[{algo}]", style="italic green")
            reason_branch = tree.add(node_text)

        if self.back:
            for back_node in self.back:
                reason_branch.add(back_node.as_rich_tree(visited))
        elif self.stage.srcs:
            for src in self.stage.srcs:
                src_text = src.compact_repr(self.stage.src_context, use_name=True)
                reason_branch.add(Text(src_text, style="green"))
        else:
            reason_branch.add(Text("...no sources...", style="bold red"))

        return tree


class MutationStage:
    def __init__(
        self,
        algorithm: SolverAlgorithm | str,
        iteration: int,
        print_context: ParameterOperatable.ReprContext,
        transformations: Transformations,
    ):
        self.algorithm = algorithm
        self.iteration = iteration
        self.transformations = transformations
        self.input_print_context = print_context
        self.input_operables = GraphFunctions(*self.input_graphs).nodes_of_type(
            ParameterOperatable
        )

    @property
    def output_graphs(self) -> list[Graph]:
        # It's enough to check for mutation graphs and not created ones
        # because the created ones always connect to graphs of the mutated ones
        # else they will be lost anyway
        return get_graphs(
            chain(
                self.transformations.mutated.values(),
                self.transformations.created,
            )
        )

    @property
    def input_graphs(self) -> list[Graph]:
        return get_graphs(self.transformations.mutated.keys())

    @property
    @once
    def output_operables(self) -> set[ParameterOperatable]:
        return GraphFunctions(*self.output_graphs).nodes_of_type(ParameterOperatable)

    @staticmethod
    def identity(
        *graphs: Graph,
        algorithm: SolverAlgorithm | str = "identity",
        iteration: int = 0,
        print_context: ParameterOperatable.ReprContext,
    ) -> "MutationStage":
        return MutationStage(
            algorithm=algorithm,
            iteration=iteration,
            print_context=print_context,
            transformations=Transformations.identity(
                *graphs, input_print_context=print_context
            ),
        )

    @property
    @once
    def is_identity(self) -> bool:
        return self.transformations.is_identity

    def as_identity(self, iteration: int = 0) -> "MutationStage":
        return MutationStage(
            algorithm="identity",
            iteration=iteration,
            print_context=self.input_print_context,
            transformations=Transformations.identity(
                *self.output_graphs, input_print_context=self.output_print_context
            ),
        )

    def print_graph_contents(
        self,
        type_filter: type[ParameterOperatable] = ParameterOperatable,
        log: Callable[[str], None] = logger.debug,
    ):
        for i, g in enumerate(self.output_graphs):
            pre_nodes = GraphFunctions(g).nodes_of_type(type_filter)
            if SHOW_SS_IS:
                nodes = pre_nodes
            else:
                nodes = [
                    n
                    for n in pre_nodes
                    if not (
                        MutatorUtils.is_alias_is_literal(n)
                        or MutatorUtils.is_subset_literal(n)
                    )
                ]
            out = ""
            node_by_depth = groupby(nodes, key=ParameterOperatable.get_depth)
            for depth, dnodes in sorted(node_by_depth.items(), key=lambda t: t[0]):
                out += f"\n  --Depth {depth}--"
                for n in dnodes:
                    out += f"\n      {n.compact_repr(self.output_print_context)}"

            if not nodes:
                continue
            log(f"|Graph {i}|={len(nodes)}/{len(pre_nodes)} [{out}\n]")

    def map_forward(self, param: ParameterOperatable) -> ParameterOperatable | None:
        if self.is_identity:
            return param
        return self.transformations.mutated.get(param)

    @property
    # FIXME not sure why but this breaks stuff, but is very necessary for speed
    @once
    def backwards_mapping(self) -> dict[ParameterOperatable, list[ParameterOperatable]]:
        return invert_dict(self.transformations.mutated)

    def map_backward(self, param: ParameterOperatable) -> list[ParameterOperatable]:
        if self.is_identity:
            return [param]
        return self.backwards_mapping.get(param, [])

    @property
    def output_print_context(self) -> ParameterOperatable.ReprContext:
        if not self.transformations:
            return self.input_print_context
        return self.transformations.output_print_context

    def print_mutation_table(self):
        if not self.transformations:
            return
        if not self.transformations.mutated:
            return

        if getattr(sys, "gettrace", lambda: None)():
            log = print
        else:
            log = logger.debug
            if not logger.isEnabledFor(logging.DEBUG):
                return

        context_old = self.input_print_context
        context_new = self.output_print_context

        created_ops = self.transformations.created

        rows: list[tuple[str, str]] = []

        for op, from_ops in created_ops.items():
            key = "new"
            key_from_ops = " \n  ".join(o.compact_repr(context_old) for o in from_ops)
            key_from_ops = f"  {key_from_ops}"
            value = op.compact_repr(context_new)
            if MutatorUtils.is_alias_is_literal(op) or MutatorUtils.is_subset_literal(
                op
            ):
                expr = next(iter(op.operatable_operands))
                lit = next(iter(op.get_operand_literals().values()))
                if not SHOW_SS_IS and expr in created_ops:
                    continue
                alias_type = "alias" if isinstance(op, Is) else "subset"
                key = f"new_{alias_type}\n{lit}"
                value = expr.compact_repr(context_new)
            if key_from_ops:
                key = f"{key} from\n{key_from_ops}"
            rows.append((key, value))

        terminated = self.transformations.terminated.difference(created_ops)
        for op in terminated:
            rows.append(("terminated", op.compact_repr(context_new)))

        copied = self.transformations.copied
        printed = set()

        for s, d in self.transformations.mutated.items():
            if not VERBOSE_TABLE:
                if s in copied:
                    continue

                # for no-op mutations (non dirty)
                if s is d:
                    continue

            old = s.compact_repr(context_old)
            new = d.compact_repr(context_new)
            if VERBOSE_TABLE:
                old += "\n\n" + repr(s)
                new += "\n\n" + repr(d)
            if old == new:
                continue
            if (
                isinstance(s, ConstrainableExpression)
                and new.replace("✓", "") == old.replace("✓", "")
                and d.try_get_literal() != s.try_get_literal()
                and new.count("✓") == old.count("✓") + 1
            ):
                # done by proven/disproven
                # TODO disproven
                continue

            printed.add(s)
            rows.append((old, new))

        merged = groupby(self.transformations.mutated.items(), key=lambda t: t[1])
        non_single_merge = {k: v for k, v in merged.items() if len(v) > 1}
        for d, sds in non_single_merge.items():
            for s, _ in sds:
                if s is d:
                    continue
                if s in printed:
                    continue
                old = s.compact_repr(context_old)
                new = d.compact_repr(context_new)
                # already printed above
                if old != new:
                    continue
                if VERBOSE_TABLE:
                    old += "\n\n" + repr(s)
                rows.append((old, "merged"))

        for s in self.transformations.removed:
            old = s.compact_repr(context_old)
            if VERBOSE_TABLE:
                old += "\n\n" + repr(s)
            rows.append((old, "removed"))

        if rows:
            rows_unique = Counter(rows)
            rows_sorted = sorted(rows_unique.items(), key=lambda t: t[0])
            table = Table(
                title="Mutations",
                show_lines=True,
            )
            track_count = any(c > 1 for c in rows_unique.values())
            if track_count:
                table.add_column("x")
            table.add_column("Before/Created By")
            table.add_column("After")
            for row, count in rows_sorted:
                count_str = "" if count == 1 else f"{count}x"
                if track_count:
                    table.add_row(count_str, *row)
                else:
                    table.add_row(*row)

            log(rich_to_string(table))

    def get_traceback_stage(self, param: ParameterOperatable) -> Traceback.Stage:
        # FIXME reenable
        # assert param in self.output_operables
        dst = param
        algo = (
            self.algorithm if isinstance(self.algorithm, str) else self.algorithm.name
        )
        related = []

        if self.is_identity:
            srcs = [param]
            reason = Traceback.Type.NOOP
        elif param in self.input_operables:
            srcs = [param]
            reason = Traceback.Type.PASSTHROUGH
        elif param in self.transformations.created:
            srcs = self.transformations.created[param]
            reason = Traceback.Type.CREATED
        elif param in self.transformations.soft_replaced:
            srcs = [
                k for k, v in self.transformations.soft_replaced.items() if v is param
            ]
            reason = Traceback.Type.SOFT_REPLACED
        else:
            origins = self.map_backward(param)
            # TODO remove (when backwards_mapping @once cache is fixed)
            assert not duplicates(origins, id)
            srcs = origins
            if len(origins) == 1:
                origin = origins[0]
                if origin in self.transformations.copied:
                    new_constraints = self.transformations.get_new_constraints(origin)
                    if new_constraints:
                        reason = Traceback.Type.CONSTRAINED
                        related_ = [
                            self.get_traceback_stage(e) for e in new_constraints
                        ]
                        for r in related_:
                            for r_s in r.srcs:
                                if r_s not in srcs:
                                    srcs.append(r_s)

                        # related.extend(related_)
                    else:
                        reason = Traceback.Type.COPIED
                else:
                    reason = Traceback.Type.MUTATED
            else:
                reason = Traceback.Type.MERGED

        return Traceback.Stage(
            srcs=srcs,
            dst=dst,
            reason=reason,
            related=related,
            algo=algo,
            src_context=self.input_print_context,
            dst_context=self.output_print_context,
        )


class MutationMap:
    @dataclass
    class LookupResult:
        maps_to: ParameterOperatable | None = None
        removed: bool = False

    def __init__(self, *stages: MutationStage):
        if not stages:
            raise ValueError("needs at least one stage")
        self.mutation_stages: list[MutationStage] = list(stages)

    @property
    @once
    def non_identity_stages(self) -> list[MutationStage]:
        return [m for m in self.mutation_stages if not m.is_identity]

    def map_forward(
        self, param: ParameterOperatable, seek_start: bool = False
    ) -> LookupResult:
        """
        return mapped param, True if removed or False if not mapped
        """
        assert isinstance(param, ParameterOperatable)
        is_root = param.get_parent() is not None

        if not self.non_identity_stages:
            out = self.first_stage.map_forward(param)
            if out is None and is_root:
                raise KeyErrorNotFound(
                    f"Looking for root parameter not in graph: {param}"
                )
            return MutationMap.LookupResult(maps_to=out)

        chain_end: ParameterOperatable = param
        if seek_start:
            first_stage = first(
                (
                    i
                    for i, m in enumerate(self.non_identity_stages)
                    if chain_end in m.input_operables
                ),
                None,
            )
            if first_stage is None:
                return MutationMap.LookupResult()
        else:
            first_stage = 0

        for m in self.non_identity_stages[first_stage:]:
            maps_to = m.map_forward(chain_end)
            if maps_to is None:
                is_start = param is chain_end
                assert not is_root or is_start, (
                    "should never remove root parameters"
                    f" chain_end {param} -> {chain_end} interrupted at"
                    f" {m.algorithm}:{m.iteration}"
                )
                if is_root and is_start:
                    raise KeyErrorNotFound(
                        f"Looking for root parameter not in graph: {param}"
                    )
                return MutationMap.LookupResult(removed=chain_end is not param)
            chain_end = maps_to
        return MutationMap.LookupResult(maps_to=chain_end)

    def map_backward(
        self, param: ParameterOperatable, only_full: bool = True
    ) -> list[ParameterOperatable]:
        chain_fronts = [param]
        collected = []

        for m in reversed(self.mutation_stages):
            next_fronts = []
            for chain_front in chain_fronts:
                maps_to = m.map_backward(chain_front)
                next_fronts.extend(maps_to)
            chain_fronts = next_fronts
            collected.extend(next_fronts)

        if only_full:
            return next_fronts

        return collected

    @property
    def compressed_mapping_forwards(self) -> dict[ParameterOperatable, LookupResult]:
        return {
            start: self.map_forward(start, seek_start=False)
            for start in self.input_operables
        }

    @property
    def compressed_mapping_forwards_complete(
        self,
    ) -> dict[ParameterOperatable, ParameterOperatable]:
        return {
            k: v.maps_to
            for k, v in self.compressed_mapping_forwards.items()
            if v.maps_to is not None
        }

    @property
    @once
    def compressed_mapping_backwards(
        self,
    ) -> dict[ParameterOperatable, list[ParameterOperatable]]:
        return {
            end: self.map_backward(end, only_full=True) for end in self.output_operables
        }

    def is_removed(self, param: ParameterOperatable) -> bool:
        return self.map_forward(param) is False

    def is_mapped(self, p: ParameterOperatable) -> bool:
        return self.map_forward(p) is not False

    def try_get_literal(
        self,
        param: ParameterOperatable,
        allow_subset: bool = False,
        domain_default: bool = False,
    ) -> SolverLiteral | None:
        def _default():
            if not domain_default:
                return None
            if not isinstance(param, Parameter):
                raise ValueError("domain_default only supported for parameters")
            return param.domain_set()

        maps_to = self.map_forward(param).maps_to
        if not isinstance(maps_to, ParameterOperatable):
            return _default()
        lit = ParameterOperatable.try_extract_literal(
            maps_to, allow_subset=allow_subset
        )
        if lit is None:
            return _default()
        lit = as_lit(lit)
        param_units = HasUnit.get_units_or_dimensionless(param)
        if isinstance(lit, Quantity_Set) and not lit.units.is_compatible_with(
            param_units
        ):
            fac = quantity(1, param_units)
            return lit * fac / fac.to_base_units().m
        return lit

    def __repr__(self) -> str:
        return f"ReprMap({str(self)})"

    def __str__(self) -> str:
        return (
            f"|stages|={len(self.mutation_stages)}"
            f", |graphs|={len(self.output_graphs)}"
            f", |V|={len(self.last_stage.output_operables)}"
        )

    @staticmethod
    def identity(
        *graphs: Graph,
        algorithm: SolverAlgorithm | str = "identity",
        iteration: int = 0,
        print_context: ParameterOperatable.ReprContext | None = None,
    ) -> "MutationMap":
        return MutationMap(
            MutationStage.identity(
                *graphs,
                algorithm=algorithm,
                iteration=iteration,
                print_context=print_context or ParameterOperatable.ReprContext(),
            )
        )

    def extend(self, *changes: MutationStage) -> "MutationMap":
        return MutationMap(*self.mutation_stages, *changes)

    @property
    def last_stage(self) -> MutationStage:
        return self.mutation_stages[-1]

    @property
    def output_graphs(self) -> list[Graph]:
        return self.last_stage.output_graphs

    @property
    def output_operables(self) -> set[ParameterOperatable]:
        return self.last_stage.output_operables

    @property
    def first_stage(self) -> MutationStage:
        return self.mutation_stages[0]

    @property
    def input_graphs(self) -> list[Graph]:
        return self.first_stage.input_graphs

    @property
    def input_operables(self) -> set[ParameterOperatable]:
        return self.first_stage.input_operables

    @property
    def output_print_context(self) -> ParameterOperatable.ReprContext:
        return self.last_stage.output_print_context

    @property
    def input_print_context(self) -> ParameterOperatable.ReprContext:
        return self.first_stage.input_print_context

    def get_iteration_mutation(self, algo: SolverAlgorithm) -> "MutationMap | None":
        last = first(
            (
                i
                for i, m in reversed(list(enumerate(self.mutation_stages)))
                if m.algorithm is algo
            ),
            None,
        )
        if last is None:
            return None
        return self.submap(start=last)

    def submap(self, start: int = 0) -> "MutationMap":
        return MutationMap(*self.mutation_stages[start:])

    def print_name_mappings(self, log: Callable[[str], None] = logger.debug):
        table = Table(title="Name mappings", show_lines=True)
        table.add_column("Variable name")
        table.add_column("Node name")

        for p in sorted(
            GraphFunctions(*self.input_graphs).nodes_of_type(Parameter),
            key=Parameter.get_full_name,
        ):
            table.add_row(p.compact_repr(self.input_print_context), p.get_full_name())

        if table.rows:
            log(rich_to_string(table))

    def get_traceback(self, param: ParameterOperatable) -> Traceback:
        start = self.last_stage.get_traceback_stage(param)
        out = Traceback(stage=start)
        deepest = [out]
        for m in reversed(self.mutation_stages[:-1]):
            new_deepest = []
            for tb in deepest:
                for op in tb.stage.srcs:
                    branch = m.get_traceback_stage(op)
                    new_tb = Traceback(stage=branch)
                    new_deepest.append(new_tb)
                    tb.back.append(new_tb)
                    # for r in branch.related:
                    #    related_tb = Traceback(stage=r)
                    #    new_deepest.append(related_tb)
                    #    tb.back.append(related_tb)
            deepest = new_deepest
        return out

    @property
    @once
    def has_merged(
        self,
    ) -> dict[ParameterOperatable, list[ParameterOperatable]]:
        mapping = self.compressed_mapping_backwards
        return {k: v for k, v in mapping.items() if len(v) > 1}

    @property
    @once
    def non_trivial_mutated_expressions(self) -> set[CanonicalExpression]:
        # TODO make faster, compact repr is a pretty bad one
        # consider congruence instead, but be careful since not in same graph space
        out = {
            v
            for v, ks in self.compressed_mapping_backwards.items()
            if isinstance(v, CanonicalExpression)
            # if all merged changed, else covered by merged
            and all(
                isinstance(k, Expression)
                and k is not v
                and k.compact_repr() != v.compact_repr()
                for k in ks
            )
        }
        return out


@dataclass
class AlgoResult:
    mutation_stage: MutationStage
    dirty: bool


class Mutator:
    # Algorithm Interface --------------------------------------------------------------

    def _mutate[T: ParameterOperatable](self, po: ParameterOperatable, new_po: T) -> T:
        """
        Low-level mutation function, you are on your own.
        Consider using mutate_parameter or mutate_expression instead.
        """
        if self.has_been_mutated(po):
            if self.get_mutated(po) is not new_po:
                raise ValueError(f"already mutated to: {self.get_mutated(po)}")

        if self.is_removed(po):
            raise ValueError("Object marked removed")

        self.transformations.mutated[po] = new_po
        return new_po

    def _override_repr(self, po: ParameterOperatable, new_po: ParameterOperatable):
        """
        Do not use this if you don't understand the consequences.
        Honestly I don't.
        """
        # TODO not sure this is the best way to handle ghost exprs
        if po in self.transformations.mutated:
            self.transformations.created[self.transformations.mutated[po]] = [po]

        self.transformations.mutated[po] = new_po

    def mutate_parameter(
        self,
        param: Parameter,
        units: Unit | Quantity | None = None,
        domain: Domain | None = None,
        soft_set: Quantity_Interval_Disjoint | Quantity_Interval | None = None,
        within: Quantity_Interval_Disjoint | Quantity_Interval | None = None,
        guess: Quantity | int | float | None = None,
        tolerance_guess: float | None = None,
        likely_constrained: bool | None = None,
        override_within: bool = False,
    ) -> Parameter:
        if param in self.transformations.mutated:
            out = self.get_mutated(param)
            assert isinstance(out, Parameter)
            assert out.units == units
            assert out.domain == domain
            assert out.soft_set == soft_set
            assert out.guess == guess
            assert out.tolerance_guess == tolerance_guess
            assert out.likely_constrained == likely_constrained
            return out

        new_param = Parameter(
            units=units if units is not None else param.units,
            within=within if override_within else param.within,
            domain=domain if domain is not None else param.domain,
            soft_set=soft_set if soft_set is not None else param.soft_set,
            guess=guess if guess is not None else param.guess,
            tolerance_guess=tolerance_guess
            if tolerance_guess is not None
            else param.tolerance_guess,
            likely_constrained=likely_constrained
            if likely_constrained is not None
            else param.likely_constrained,
        )

        return self._mutate(param, new_param)

    def _create_expression[T: Expression](
        self,
        expr_factory: type[T],
        *operands: SolverAllExtended,
        non_operands: Any = None,
        constrain: bool = False,
    ) -> T:
        new_operands = [
            self.get_copy(
                op,
                accept_soft=not (expr_factory in [Is, IsSubset] and constrain),
            )
            for op in operands
        ]
        if non_operands is not None:
            new_expr = expr_factory(*new_operands, non_operands=non_operands)  # type: ignore
        else:
            new_expr = expr_factory(*new_operands)

        if constrain and isinstance(new_expr, ConstrainableExpression):
            new_expr.constrained = True
            # TODO this is better, but ends up in inf loop
            # self.constrain(new_expr)

        for op in new_operands:
            if isinstance(op, ParameterOperatable):
                assert (
                    op.get_graph() == new_expr.get_graph()
                ), f"Graph mismatch: {op.get_graph()} != {new_expr.get_graph()}"

        return new_expr

    def mutate_expression(
        self,
        expr: Expression,
        operands: Iterable[SolverAllExtended] | None = None,
        expression_factory: type[Expression] | None = None,
        soft_mutate: type[Is] | type[IsSubset] | None = None,
        ignore_existing: bool = False,
        from_ops: Sequence[ParameterOperatable] | None = None,
    ) -> CanonicalExpression:
        if expression_factory is None:
            expression_factory = type(expr)

        if operands is None:
            operands = expr.operands

        if expr in self.transformations.mutated:
            out = self.get_mutated(expr)
            assert isinstance(out, CanonicalExpression)
            # TODO more checks
            assert type(out) is expression_factory
            # still need to run soft_mutate even if expr already in repr
            if soft_mutate:
                expr = out
            else:
                return out

        if soft_mutate:
            assert issubclass(expression_factory, CanonicalExpression)
            # TODO: technically the return type is incorrect, but it's not used anywhere
            # if run with soft_mutaste
            return self.soft_mutate_expr(
                expression_factory, expr, operands, soft_mutate, from_ops=from_ops
            )  # type: ignore

        if from_ops is not None:
            raise NotImplementedError("only supported for soft_mutate")

        copy_only = expression_factory is type(expr) and operands == expr.operands
        if not copy_only and not ignore_existing:
            assert issubclass(expression_factory, CanonicalExpression)
            exists = self.utils.find_congruent_expression(
                expression_factory, *operands, allow_uncorrelated=False
            )
            if exists is not None:
                return self._mutate(expr, self.get_copy(exists))

        constrain = isinstance(expr, ConstrainableExpression) and expr.constrained
        new_expr = self._create_expression(
            expression_factory,
            *operands,
            non_operands=expr.non_operands,
            constrain=constrain,
        )

        if isinstance(expr, ConstrainableExpression):
            new_expr = cast_assert(ConstrainableExpression, new_expr)
            if self.is_predicate_terminated(expr):
                new_expr._solver_terminated = True

        return self._mutate(expr, new_expr)  # type: ignore #TODO

    def soft_replace[T: ParameterOperatable](
        self,
        current: T,
        new: ParameterOperatable,
    ) -> T:
        if self.has_been_mutated(current):
            copy = self.get_mutated(current)
            exps = copy.get_operations()  # noqa: F841
            # FIXME: reenable, but alias classes need to take this into account
            # assert all(isinstance(o, (Is, IsSubset)) and o.constrained for o in exps)

        self.transformations.soft_replaced[current] = new
        return self.get_copy(current, accept_soft=False)  # type: ignore

    def soft_mutate_expr(
        self,
        expression_factory: type[CanonicalExpression],
        expr: Expression,
        operands: Iterable[SolverAllExtended],
        soft: type[Is] | type[IsSubset],
        from_ops: Sequence[ParameterOperatable] | None = None,
    ):
        operands = list(operands)
        # Don't create A is A, lit is lit
        if expr.is_congruent_to_factory(
            expression_factory, operands, allow_uncorrelated=True
        ):
            return expr  # type: ignore

        # Avoid alias X to Op1(lit) if X is!! Op2(lit)
        congruent = {
            alias
            for alias in (
                self.utils.get_aliases(expr)
                if soft is Is
                else self.utils.get_supersets(expr)
            )
            if isinstance(alias, expression_factory)
            and alias.is_congruent_to_factory(
                expression_factory, operands, allow_uncorrelated=True
            )
        }
        if congruent:
            return next(iter(congruent))

        out = self.create_expression(
            expression_factory,
            *operands,
            from_ops=[expr, *(from_ops or [])],
            allow_uncorrelated=soft is IsSubset,
        )
        self.soft_mutate(soft, expr, out, from_ops=from_ops)
        return out

    # TODO make more use of soft_mutate for alias & ss with non-lit
    def soft_mutate(
        self,
        soft: type[Is] | type[IsSubset],
        old: ParameterOperatable,
        new: SolverAll,
        from_ops: Sequence[ParameterOperatable] | None = None,
    ):
        # filter A is A, A ss A
        if new is old:
            return
        self.create_expression(
            soft,
            old,
            new,
            constrain=True,
            from_ops=unique_ref([old] + list(from_ops or [])),
            # FIXME
            allow_uncorrelated=True,
        )

    def mutate_unpack_expression(
        self, expr: Expression, operands: list[ParameterOperatable] | None = None
    ) -> ParameterOperatable:
        """
        ```
        op(A, ...) -> A
        op!(A, ...) -> A!
        ```
        """
        unpacked = expr.operands[0] if operands is None else operands[0]
        if not isinstance(unpacked, ParameterOperatable):
            raise ValueError("Unpacked operand can't be a literal")
        out = self._mutate(expr, self.get_copy(unpacked))
        if isinstance(expr, ConstrainableExpression) and expr.constrained:
            assert isinstance(out, ConstrainableExpression)
            self.constrain(out)
        return out

    def mutator_neutralize_expressions(self, expr: Expression) -> ParameterOperatable:
        """
        '''
        op(op_inv(A), ...) -> A
        op!(op_inv(A), ...) -> A!
        '''
        """
        inner_expr = expr.operands[0]
        if not isinstance(inner_expr, Expression):
            raise ValueError("Inner operand must be an expression")
        inner_operand = inner_expr.operands[0]
        if not isinstance(inner_operand, ParameterOperatable):
            raise ValueError("Unpacked operand can't be a literal")
        out = self._mutate(expr, self.get_copy(inner_operand))
        if isinstance(out, ConstrainableExpression) and out.constrained:
            self.constrain(out)
        return out

    def mutate_expression_with_op_map(
        self,
        expr: Expression,
        operand_mutator: Callable[[int, ParameterOperatable], ParameterOperatable.All],
        expression_factory: type[CanonicalExpression] | None = None,
        ignore_existing: bool = False,
    ) -> CanonicalExpression:
        """
        operand_mutator: Only allowed to return old Graph objects
        """
        return self.mutate_expression(
            expr,
            operands=[operand_mutator(i, op) for i, op in enumerate(expr.operands)],
            expression_factory=expression_factory,
            ignore_existing=ignore_existing,
        )

    def get_copy(
        self, obj: ParameterOperatable.All, accept_soft: bool = True
    ) -> ParameterOperatable.All:
        if not isinstance(obj, ParameterOperatable):
            return obj

        if accept_soft and obj in self.transformations.soft_replaced:
            return self.transformations.soft_replaced[obj]

        if self.has_been_mutated(obj):
            return self.get_mutated(obj)

        # TODO: not sure if ok
        # if obj is new, no need to copy
        # TODO add guard to _mutate to not let new stuff be mutated
        if (
            obj in self.transformations.created
            or obj in self.transformations.mutated.values()
        ):
            return obj

        # purely for debug
        self.transformations.copied.add(obj)

        if isinstance(obj, Expression):
            return self.mutate_expression(obj)
        elif isinstance(obj, Parameter):
            return self.mutate_parameter(obj)

        assert False

    def create_expression[T: CanonicalExpression](
        self,
        expr_factory: type[T],
        *operands: SolverAll,
        check_exists: bool = True,
        from_ops: Sequence[ParameterOperatable] | None = None,
        constrain: bool = False,
        allow_uncorrelated: bool = False,
        _relay: bool = True,
    ) -> T | IsSubset | Is | SolverLiteral:
        from faebryk.core.solver.symbolic.pure_literal import (
            _exec_pure_literal_operands,
        )

        assert issubclass(expr_factory, CanonicalExpression)
        from_ops = [
            x for x in unique_ref(from_ops or []) if isinstance(x, ParameterOperatable)
        ]
        if _relay:
            if constrain and expr_factory is IsSubset:
                return self.utils.subset_to(operands[0], operands[1], from_ops=from_ops)
            if constrain and expr_factory is Is:
                return self.utils.alias_to(operands[0], operands[1], from_ops=from_ops)
            res = _exec_pure_literal_operands(expr_factory, operands)
            if res is not None:
                if constrain and res != as_lit(True):
                    raise ContradictionByLiteral(
                        "Literal is not true",
                        involved=from_ops,
                        literals=[res],
                        mutator=self,
                    )
                return res

        expr = None
        if check_exists:
            # TODO look in old & new graph
            expr = self.utils.find_congruent_expression(
                expr_factory,
                *operands,
                allow_uncorrelated=allow_uncorrelated,
            )

        if expr is None:
            expr = self._create_expression(
                expr_factory,
                *operands,
                constrain=constrain,
            )
            self.transformations.created[expr] = from_ops

        # TODO double constrain ugly
        if constrain and isinstance(expr, ConstrainableExpression):
            self.constrain(expr)

        return expr

    def remove(self, *po: ParameterOperatable):
        assert not any(
            p in self.transformations.mutated for p in po
        ), "Object already in repr_map"
        root_pos = [p for p in po if p.get_parent() is not None]
        assert not root_pos, f"should never remove root parameters: {root_pos}"
        self.transformations.removed.update(po)

    def remove_graph(self, g: Graph):
        # TODO implementing graph removal has to be more explicit
        # e.g mark as no more use, and then future mutators ignore it for the algos
        # for now at least remove expressions
        assert g in self.G
        self.remove(*GraphFunctions(g).nodes_of_type(Expression))

    def register_created_parameter(
        self, param: Parameter, from_ops: Sequence[ParameterOperatable] | None = None
    ) -> Parameter:
        self.transformations.created[param] = list(from_ops or [])
        return param

    def constrain(self, *po: ConstrainableExpression, terminate: bool = False):
        for p in po:
            p.constrain()
            self.utils.alias_to(p, as_lit(True), terminate=terminate)

    def predicate_terminate(self, pred: ConstrainableExpression):
        assert pred.constrained
        if pred._solver_terminated:
            return
        pred._solver_terminated = True
        self.transformations.terminated.add(pred)

    def predicate_reset_termination(self, pred: ConstrainableExpression):
        assert pred.constrained
        if not pred._solver_terminated:
            return
        pred._solver_terminated = False

    # Algorithm Query ------------------------------------------------------------------
    def is_predicate_terminated(self, pred: ParameterOperatable) -> bool:
        assert isinstance(pred, ConstrainableExpression)
        return pred._solver_terminated

    def nodes_of_type[T: "ParameterOperatable"](
        self,
        t: type[T] = ParameterOperatable,
        sort_by_depth: bool = False,
        created_only: bool = False,
        new_only: bool = False,
        include_terminated: bool = False,
    ) -> list[T] | set[T]:
        assert not new_only or not created_only

        if new_only:
            out = {n for n in self._new_operables if isinstance(n, t)}
        elif created_only:
            out = {n for n in self.transformations.created if isinstance(n, t)}
        else:
            out = GraphFunctions(*self.G).nodes_of_type(t)

        if not include_terminated:
            out = {
                n
                for n in out
                if not (
                    isinstance(n, ConstrainableExpression)
                    and self.is_predicate_terminated(n)
                )
            }

        if sort_by_depth:
            out = ParameterOperatable.sort_by_depth(out, ascending=True)

        return out

    def nodes_of_types(
        self,
        t: tuple[type[ParameterOperatable], ...] | UnionType,
        sort_by_depth: bool = False,
        include_terminated: bool = False,
        new_only: bool = False,
    ) -> list[ParameterOperatable] | set[ParameterOperatable]:
        if new_only:
            out = {n for n in self._new_operables if isinstance(n, t)}
        else:
            out = GraphFunctions(*self.G).nodes_of_types(t)
        out = cast(set[ParameterOperatable], out)
        if not include_terminated:
            out = {
                n
                for n in out
                if not (
                    isinstance(n, ConstrainableExpression)
                    and self.is_predicate_terminated(n)
                )
            }
        if sort_by_depth:
            out = ParameterOperatable.sort_by_depth(out, ascending=True)
        return out

    @property
    def non_copy_mutated(self) -> set[CanonicalExpression]:
        if self._mutations_since_last_iteration is None:
            return set()
        return self._mutations_since_last_iteration.non_trivial_mutated_expressions

    def get_literal_aliases(self, new_only: bool = True):
        """
        Find new ops which are Is expressions between a ParameterOperatable and a
        literal
        """

        aliases: set[CanonicalExpression]
        aliases = set(
            self.nodes_of_type(Is, new_only=new_only, include_terminated=True)
        )

        if new_only and self._mutations_since_last_iteration is not None:
            # Taking into account if op with no literal merged into a op with literal
            mapping = self._mutations_since_last_iteration.has_merged
            for new, olds in mapping.items():
                new_lit = self.utils.try_extract_literal(new)
                if new_lit is None:
                    continue
                old_lits = {self.utils.try_extract_literal(o) for o in olds}
                if old_lits == {new_lit}:
                    continue
                aliases.update(new.get_operations(Is, constrained_only=True))
            aliases.update(
                self._mutations_since_last_iteration.non_trivial_mutated_expressions
            )

        return (expr for expr in aliases if self.utils.is_alias_is_literal(expr))

    def _get_literal_subsets(self, new_only: bool = True):
        subsets: set[CanonicalExpression]
        subsets = set(
            self.nodes_of_type(IsSubset, new_only=new_only, include_terminated=True)
        )

        if new_only and self._mutations_since_last_iteration is not None:
            # Taking into account if op with no literal merged into a op with literal
            mapping = self._mutations_since_last_iteration.has_merged
            for new, olds in mapping.items():
                new_lit = self.utils.try_extract_literal(new, allow_subset=True)
                if new_lit is None:
                    continue
                old_lits = {
                    self.utils.try_extract_literal(o, allow_subset=True) for o in olds
                }
                if old_lits == {new_lit}:
                    continue
                subsets.update(new.get_operations(IsSubset, constrained_only=True))
            subsets.update(
                self._mutations_since_last_iteration.non_trivial_mutated_expressions
            )

        return (expr for expr in subsets if self.utils.is_subset_literal(expr))

    def get_literal_mappings(self, new_only: bool = True, allow_subset: bool = False):
        # TODO better exceptions

        ops = self.get_literal_aliases(new_only=new_only)
        mapping = {self.utils.get_lit_mapping_from_lit_expr(op) for op in ops}
        dupes = duplicates(mapping, lambda x: x[0], by_eq=True)
        if dupes:
            raise ContradictionByLiteral(
                "Literal contradictions",
                list(dupes.keys()),
                list(v[1] for vs in dupes.values() for v in vs),
                mutator=self,
            )
        mapping_dict = dict(mapping)

        if allow_subset:
            ops_ss = self._get_literal_subsets(new_only=new_only)
            mapping_ss = [self.utils.get_lit_mapping_from_lit_expr(op) for op in ops_ss]
            grouped_ss = groupby(mapping_ss, key=lambda t: t[0])
            for k, v in grouped_ss.items():
                ss_lits = [ss_lit for _, ss_lit in v]
                merged_ss = P_Set.intersect_all(*ss_lits)
                if merged_ss.is_empty():
                    raise ContradictionByLiteral(
                        "Empty intersection", [k], ss_lits, mutator=self
                    )
                if k in mapping_dict:
                    if not mapping_dict[k].is_subset_of(merged_ss):  # type: ignore
                        raise ContradictionByLiteral(
                            "is lit not subset of ss lits",
                            [k],
                            [mapping_dict[k], *ss_lits],
                            mutator=self,
                        )
                    continue
                mapping_dict[k] = merged_ss

        return mapping_dict

    def is_removed(self, po: ParameterOperatable) -> bool:
        return po in self.transformations.removed

    def has_been_mutated(self, po: ParameterOperatable) -> bool:
        return po in self.transformations.mutated

    def get_mutated(self, po: ParameterOperatable) -> ParameterOperatable:
        return self.transformations.mutated[po]

    # Solver Interface -----------------------------------------------------------------
    def __init__(
        self,
        mutation_map: MutationMap,
        algo: SolverAlgorithm,
        iteration: int,
        terminal: bool,
    ) -> None:
        self.algo = algo
        self.terminal = terminal
        self.mutation_map = mutation_map
        self.iteration = iteration

        self.utils = MutatorUtils(self)

        self._G: set[Graph] = set(mutation_map.output_graphs)
        self.print_context = mutation_map.output_print_context
        self._mutations_since_last_iteration = mutation_map.get_iteration_mutation(algo)

        self._starting_operables = set(self.nodes_of_type(include_terminated=True))

        self.transformations = Transformations(input_print_context=self.print_context)

    @property
    @once
    def _new_operables(self) -> set[ParameterOperatable]:
        _last_run_operables = set()
        if self._mutations_since_last_iteration is not None:
            _last_run_operables = set(
                self._mutations_since_last_iteration.compressed_mapping_forwards_complete.values()
            )
        assert _last_run_operables.issubset(self._starting_operables)
        return self._starting_operables - _last_run_operables

    @property
    def G(self) -> set[Graph]:
        # Handles C++ graph shenanigans on move
        gs = self._G
        if all(g.node_count > 0 for g in gs):
            return gs
        # Handle graph merge
        gs = get_graphs(self._starting_operables)
        self._G = set(gs)
        return self._G

    def _run(self):
        self.algo(self)

    def _copy_unmutated(
        self,
    ):
        _touched_graphs = self.transformations.touched_graphs
        touched = self.transformations.mutated.keys() | self.transformations.removed

        # TODO might not need to sort
        other_param_op = ParameterOperatable.sort_by_depth(
            (
                GraphFunctions(*_touched_graphs).nodes_of_type(ParameterOperatable)
                - touched
            ),
            ascending=True,
        )
        for p in other_param_op:
            self.get_copy(p)

        # optimization: if just new_ops, no need to copy
        # pass through untouched graphs
        untouched_graphs = self.G - _touched_graphs
        for p in GraphFunctions(*untouched_graphs).nodes_of_type(ParameterOperatable):
            self.transformations.mutated[p] = p

    def check_no_illegal_mutations(self):
        # TODO should only run during dev

        # Check modifications to original graph
        post_mut_nodes = set(self.nodes_of_type(include_terminated=True))
        removed = self._starting_operables.difference(
            post_mut_nodes, self.transformations.removed
        )
        added = post_mut_nodes.difference(
            self._starting_operables, self.transformations.created
        )
        removed_compact = [op.compact_repr(self.print_context) for op in removed]
        added_compact = [op.compact_repr(self.print_context) for op in added]
        assert not removed, (
            f"Mutator {self.G, self.algo.name} untracked removed "
            f"{indented_container(removed_compact)}"
        )
        assert not added, (
            f"Mutator {self.G, self.algo.name} untracked added "
            f"{indented_container(added_compact)}"
        )

        # don't need to check original graph, done above seperately
        all_new_graphs = get_graphs(self.transformations.mutated.values())
        all_new_params = {
            op
            for g in all_new_graphs
            for op in GraphFunctions(g).nodes_of_type(ParameterOperatable)
        }
        non_registered = all_new_params.difference(
            self.transformations.created, self.transformations.mutated.values()
        )
        if non_registered:
            compact = (op.compact_repr(self.print_context) for op in non_registered)
            graphs = get_graphs(non_registered)
            # FIXME: this is currently hit during legitimate build
            with downgrade(AssertionError, logger=logger, to_level=logging.DEBUG):
                assert False, (
                    f"Mutator {self.G} has non-registered new ops: "
                    f"{indented_container(compact)}."
                    f"{indented_container(graphs)}"
                )

    def close(self) -> AlgoResult:
        if not self.transformations.dirty:
            return AlgoResult(
                mutation_stage=MutationStage.identity(
                    *self.mutation_map.output_graphs,
                    algorithm=self.algo,
                    iteration=self.iteration,
                    print_context=self.print_context,
                ),
                dirty=False,
            )

        touched_pre_copy = self.transformations.touched_graphs
        self.check_no_illegal_mutations()
        self._copy_unmutated()
        stage = MutationStage(
            algorithm=self.algo,
            iteration=self.iteration,
            transformations=self.transformations,
            print_context=self.print_context,
        )

        # Check if original graphs ended up in result
        # allowed if no copy was needed for graph
        assert not (touched_pre_copy & set(stage.output_graphs))

        return AlgoResult(mutation_stage=stage, dirty=True)

    def run(self):
        self._run()
        return self.close()

    # Debug Interface ------------------------------------------------------------------
    def __repr__(self) -> str:
        return f"Mutator('{self.algo.name}' {self.transformations})"



import os
import shutil
import sys
from dataclasses import dataclass
from pathlib import Path
from subprocess import CalledProcessError

import git
import pathvalidate
import pytest

from faebryk.libs.util import repo_root as _repo_root
from faebryk.libs.util import robustly_rm_dir, run_live


class CloneError(Exception):
    """Failed to clone the repository."""


class InstallError(Exception):
    """Failed to install the project."""


class BuildError(Exception):
    """Failed to build the project."""


ENABLE_PROFILING = False


def build_project(prj_path: Path, request: pytest.FixtureRequest):
    """Generically "build" the project."""
    friendly_node_name = pathvalidate.sanitize_filename(str(request.node.name))
    artifact_dir = _repo_root() / "artifacts"

    try:
        ato_build_args = [
            "--keep-picked-parts",
            "--keep-net-names",
        ]
        ato_build_env = {"NONINTERACTIVE": "1"}

        profile = []
        if ENABLE_PROFILING:
            profile_path = artifact_dir / (friendly_node_name + "-profile.html")
            profile = [
                "-m",
                "pyinstrument",
                "--html",
                "-o",
                profile_path,
                "-i",
                "0.01",
            ]
        run_live(
            [
                sys.executable,
                *profile,
                "-m",
                "atopile",
                "-v",
                "build",
                *ato_build_args,
            ],
            env={**os.environ, **ato_build_env},
            cwd=prj_path,
            stdout=print,
            stderr=print,
        )
    except CalledProcessError as ex:
        artifact_path = artifact_dir / friendly_node_name
        if artifact_path.exists():
            robustly_rm_dir(artifact_path)
        artifact_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(prj_path, artifact_path, ignore=shutil.ignore_patterns(".git"))

        # Translate the error message to clearly distinguish from clone errors
        raise BuildError from ex


@dataclass
class _TestRepo:
    repo_uri: str
    xfail_reason: str | None = None
    multipackage: str | None = None

    def __post_init__(self):
        if not self.repo_uri.startswith("https") and not self.repo_uri.startswith(
            "ssh"
        ):
            self.repo_uri = f"https://github.com/{self.repo_uri}"

    def xfail(self, reason: str):
        self.xfail_reason = reason
        return self


REPOS = [
    _TestRepo("atopile/spin-servo-drive").xfail("Needs upgrading"),
    _TestRepo("atopile/packages", multipackage="packages"),
]


@pytest.mark.slow
@pytest.mark.regression
@pytest.mark.parametrize(
    "test_cfg",
    REPOS,
    ids=lambda x: x.repo_uri,
)
def test_projects(
    test_cfg: _TestRepo,
    tmp_path: Path,
    request: pytest.FixtureRequest,
):
    repo_uri = test_cfg.repo_uri
    xfail_reason = test_cfg.xfail_reason
    multipackage = test_cfg.multipackage

    # Clone the repository
    # Using gh to use user credentials if run locally
    try:
        run_live(
            ["gh", "repo", "clone", repo_uri, "project", "--", "--depth", "1"],
            cwd=tmp_path,
            stdout=print,
            stderr=print,
        )
    except CalledProcessError as ex:
        raise CloneError from ex

    prj_path = tmp_path / "project"

    def run_project(path: Path):
        # Generically "install" the project
        try:
            ato_dir = path / ".ato"
            if ato_dir.exists():
                robustly_rm_dir(ato_dir)
            run_live(
                [sys.executable, "-m", "atopile", "-v", "sync"],
                env={**os.environ, "NONINTERACTIVE": "1"},
                cwd=path,
                stdout=print,
                stderr=print,
            )
        except CalledProcessError as ex:
            # Translate the error message to clearly distinguish from clone errors
            raise InstallError from ex

        build_project(path, request=request)

    try:
        if multipackage:
            subpath = prj_path / multipackage
            for subdir in subpath.iterdir():
                if subdir.is_dir():
                    run_project(subdir)
        else:
            run_project(prj_path)
    except (InstallError, BuildError):
        if xfail_reason:
            pytest.xfail(xfail_reason)
        else:
            raise

    repo = git.Repo(prj_path)
    diff = repo.index.diff(None)
    # Check if diff exists and if any item in the diff represents
    # a change to a KiCad PCB file
    if diff and any(
        item.a_path is not None and item.a_path.endswith(".kicad_pcb") for item in diff
    ):
        pytest.xfail("can't build with --frozen yet")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L


class USB_Type_C_Receptacle_16_pin(Module):
    # interfaces

    # TODO make arrays?
    cc1: F.Electrical
    cc2: F.Electrical
    sbu1: F.Electrical
    sbu2: F.Electrical
    shield: F.Electrical
    # power
    power: F.ElectricPower
    # ds: p, n
    d: F.USB2_0_IF.Data

    @L.rt_field
    def attach_to_footprint(self):
        vbus = self.power.hv
        gnd = self.power.lv

        return F.can_attach_to_footprint_via_pinmap(
            {
                "A1": gnd,
                "A4": vbus,
                "A5": self.cc1,
                "A6": self.d.p.line,
                "A7": self.d.n.line,
                "A8": self.sbu1,
                "A9": vbus,
                "A12": gnd,
                "B1": gnd,
                "B4": vbus,
                "B5": self.cc2,
                "B6": self.d.p.line,
                "B7": self.d.n.line,
                "B8": self.sbu2,
                "B9": vbus,
                "B12": gnd,
                "0": self.shield,
                "1": self.shield,
                "2": self.shield,
                "3": self.shield,
            }
        )

    @L.rt_field
    def pin_association_heuristic(self):
        vbus = self.power.hv
        gnd = self.power.lv

        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                gnd: ["GND"],
                vbus: ["VBUS"],
                self.cc1: ["CC1"],
                self.cc2: ["CC2"],
                self.d.p.line: ["DP1", "DP2"],
                self.d.n.line: ["DN1", "DN2"],
                self.sbu1: ["SBU1"],
                self.sbu2: ["SBU2"],
                self.shield: ["SHELL"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )
